/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/testbed/commands.c 
 * @brief the commands available in the testbed
 * @author Ronaldo Alves Ferreira
 * @author Christian Grothoff
 * @author Murali Krishan Ramanathan
 *
 * Todo:
 * - implement more messages
 * - test
 **/

#include "testbed.h"
#include "commands.h"
#include "socket.h"



/**
 * @brief struct keeping per-peer information for the testbed
 */
typedef struct {
  /** IP address of the peer */
  IPaddr	     ip;
  /** CS port of the peer */
  unsigned short     port;
  /** string describing the peer address */
  char	* ips;
  /** socket to communicate with the peer */
  GNUNET_TCP_SOCKET sock;
  /** HELO message identifying the peer in the network */
  HELO_Message	    *helo;
} NODE_INFO;

/**
 * List of nodes known to the testbed.
 */
static NODE_INFO * nodes = NULL;

/*
 * Number of known nodes, size of the nodes array.
 */
static int nnodes = 0;

/*
 * Number of active connections
 */
static int connections;


/**
 * Add a node to the configuration.
 */
static int addNode(int argc,
		    char * argv[]) {
  int currindex;
  TESTBED_ADD_PEER_MESSAGE * hdr;
  char ip_aux[30];
  TESTBED_CS_MESSAGE *msg;
  char * line = argv[0];
  char * s = line;
    
  PRINTF("Adding node: %s\n", s);
    
  /* FIXME: the code below needs to be improved. */
  while (*s != ':') s++;
  *s++ = '\0';
  strcpy(ip_aux, line);
  
  currindex = nnodes;
  GROW(nodes,
       nnodes,
       nnodes+1);
  nodes[currindex].ips = STRDUP(ip_aux);
  nodes[currindex].port = atoi(s);
  inet_aton(ip_aux, &nodes[currindex].ip.addr);
  PRINTF("IP: %s Port:%d\n", 
	 ip_aux, 
	 nodes[currindex].port);
  
  if (SYSERR != initGNUnetClientSocket(nodes[currindex].port,
				       ip_aux, 
				       &nodes[currindex].sock)) {     
    TESTBED_GET_HELO_MESSAGE *phelo;
    
    msg = MALLOC(sizeof(TESTBED_GET_HELO_MESSAGE));
    msg->header.size = htons(sizeof(TESTBED_GET_HELO_MESSAGE));
    msg->header.tcpType = htons(TESTBED_CS_PROTO_REQUEST);
    msg->msgType = htonl(TESTBED_GET_HELO);
    phelo = (TESTBED_GET_HELO_MESSAGE *)msg->data;
    phelo->proto = htons(TCP_PROTOCOL_NUMBER);
    if (SYSERR == writeToSocket(&nodes[currindex].sock, 
				&msg->header)) {
      PRINTF("Could not send message to node %s.  Trying other available peers.\n", 
	     nodes[currindex].ips);
      FREE(msg);
      FREE(nodes[currindex].ips);
      GROW(nodes, 
	   nnodes,
	   nnodes-1);
      return -1;
    }
    FREE(msg);
    
    PRINTF("writeToSocket returned successfully\n");
    hdr = NULL;      
    if (OK == readFromSocket(&nodes[currindex].sock, 
			     (CS_HEADER**)&hdr)) {
      nodes[currindex].helo = MALLOC(HELO_Message_size(&hdr->helo));
      memcpy(nodes[currindex].helo, 
	     &hdr->helo, 
	     HELO_Message_size(&hdr->helo));
      PRINTF("HELO message received\n"\
	     "\tprotocol: %d\n"\
	     "\tMTU: %u\n",
	     ntohs(hdr->helo.protocol), 
	     (unsigned int) ntohl(hdr->helo.MTU));
      FREE(hdr);
    } else {
      LOG(LOG_WARNING,
	  "WARNING: failed to receive HELO from peer!\n");
      FREE(nodes[currindex].ips);
      GROW(nodes, 
	   nnodes,
	   nnodes-1);
      return -1;
    }
  }
  return 0;
}


/* */
static int sendMessage(unsigned int msgType, 
		       int dst, 
		       unsigned int argSize,
		       void * arg) {
  TESTBED_CS_MESSAGE * msg;
  int msgsz;

  if (dst >= nnodes || dst < 0) {
    PRINTF("Invalid destination node %d\n", dst);
    return SYSERR;
  }
  /* Nodes are valid. Send message to delete connection. */      
  msgsz = sizeof(TESTBED_CS_MESSAGE)+argSize;
  msg = MALLOC(msgsz);
  msg->header.size = htons(msgsz);
  msg->header.tcpType = htons(TESTBED_CS_PROTO_REQUEST);
  msg->msgType = htonl(msgType);
  memcpy(&msg->data[0],
	 arg,
	 argSize);
  PRINTF("Sending message to node %s\n", 
	 nodes[dst].ips);
  if (SYSERR == writeToSocket(&nodes[dst].sock, 
			      &msg->header))
    PRINTF("Could not send message to node %s\n",
	   nodes[dst].ips);
  else
    PRINTF("writeToSocket returned successfully\n"); 
  FREE(msg);
  return OK;
}

/* */
static int delConnection(int argc,
			  char * argv[]) {
  int src, dst;
  char *s = argv[0];

  sscanf (s, "%d %d", &src, &dst);
  PRINTF("del %d->%d\n", src, dst);
  if (OK == sendMessage(TESTBED_DEL_PEER, 
			src,
			sizeof(HostIdentity),
			&nodes[dst].helo->senderIdentity)) {
    /* FIXME: check return value from peer! */
    connections--;
  }
  return 0;
}

/* */
static int addConnection(int argc,
			 char * argv[]) {
  int src, dst;
  char *s = argv[0];
  
  sscanf (s, "%d %d", &src, &dst);
  PRINTF("add %d->%d\n", src, dst);
  if (OK == sendMessage(TESTBED_ADD_PEER,
			src,
			dst,
			0)) {
    /* FIXME: check a return value from
       the peer! */
    connections++;
  }
  return 0;
}

/* */
static int setTrust(int argc,
		    char * argv[]) {
  int src, dst, value;
  int confirmation;
  char *s = argv[0];
  
  sscanf (s, "%d %d %d", &src, &dst, &value);
  PRINTF("set-trust %d->%d %d\n", src, dst, value);
  /* sendMessage(TESTBED_SET_TVALUE, dst, value, SIZE, MSG);*/

  if (OK == readTCPResult(&nodes[src].sock,
			  &confirmation)) {
    if (confirmation == OK) {
      PRINTF("Trust value between %d and %d set to %d\n", 
	     src, 
	     dst, 
	     value);
    } else {
      LOG(LOG_WARNING,
	  "WARNING: peer encountered error setting trust value.\n");
    }
  } else {
    LOG(LOG_ERROR,
	"ERROR: peer did not confirm setting trust value.\n");
  }  
  return 0;
}

/**
 * ATTN RF: I've hacked this method to show a bit more how the
 * protocol should work IMO.  You may want to model some of the others
 * accordingly. -- CG
 * Addition: I did not hack the argument parsing and verification.
 */
static int getTrust(int argc,
		    char * argv[]) {
  int   src, dst, value;
  char *s = argv[0];
  
  sscanf (s, "%d %d", &src, &dst);
  if (OK == sendMessage(TESTBED_GET_TVALUE, 
			src, 
			sizeof(HostIdentity),
			&nodes[dst].helo->senderIdentity)) {
    if (OK == readTCPResult(&nodes[src].sock,
			    &value)) {
      setConfigurationInt("GNUNET-TESTBED-VARIABLE",
			  "$0",
			  value);
      LOG(LOG_INFO,
	  "INFO: Trust value between %d and %d is %d\n", 
	  src, 
	  dst, 
	  value);
    } else {
      LOG(LOG_ERROR,
	  "ERROR: could not get trust value.\n");
      setConfigurationInt("GNUNET-TESTBED-VARIABLE",
			  "$0",
			  -1);
    }
  } else {
    LOG(LOG_ERROR,
	"ERROR: could not send request for trust value.\n");
  }
  return 0;
}

/**
 * Helper function for (un)loadModule.
 * @param type load or unload requested?
 */
static int XXloadModule(unsigned short type,
			char * cmdLine) {
  int ok;
  int dst;
  char * modulename;

  modulename = cmdLine;
  while ( ( (*modulename) != '\0') &&
	  ( (*modulename) != ' ') )
    modulename++;
  if ( *modulename != ' ') {
    PRINTF("Input error.  Format: load-module PEERID MODULENAME.\n");
    return -1;
  }
  *modulename = '\0';
  modulename++;
  if (strlen(modulename) == 0) {
    PRINTF("Input error.  You must specify a module name.\n");
    return -1;
  }
  if (1 != sscanf(cmdLine, "%d", &dst)) {
    PRINTF("Input error.  PEERID must be number.\n");
    return -1;
  }   
  ok = sendMessage(type,
		   dst,
		   strlen(modulename),
		   modulename);
  if (ok == OK) {
    if (OK == readTCPResult(&nodes[dst].sock,
			    &ok)) {
      if (ok != OK) {
	LOG(LOG_WARNING,
	    "WARNING: peer had error loading module.\n");
	return -1;
      }
    } else {
      LOG(LOG_ERROR,
	  "ERROR: peer did not send confirmation message for loadModule.\n");
      return -1;
    }
    return 0;
  } else {
    LOG(LOG_ERROR,
	"ERROR: could not send loadModule message.\n");
    return -1;
  }
}

static int loadModule(int argc,
		      char * argv[]) {
  return XXloadModule(TESTBED_LOAD_MODULE,
		      argv[0]);
}
static int unloadModule(int argc,
			char * argv[]) {
  return XXloadModule(TESTBED_UNLOAD_MODULE,
		      argv[0]);
}

/* */
static int setBW(int argc,
		 char * argv[]) {
    int src, dst, in, out;
    char *s = argv[0];
    
    sscanf(s,
	   "%d %d %d %d",
	   &src, 
	   &dst, 
	   &in, 
	   &out);
    PRINTF("set-trust %d->%d in=%d out=%d\n", 
	   src, 
	   dst,
	   in, 
	   out);
    /*    if (OK == sendMessage(TESTBED_SET_IN_BW, 
			  src,
			  dst, 
			  in))
      sendMessage(TESTBED_SET_OUT_BW, 
		  src, 
		  dst, 
		  out);
    */
    return 0;
}

/* */
static int setLoss(int argc,
		   char * argv[]) {
  int src, dst, value;
  char *s = argv[0];
  
  sscanf (s, "%d %d %d",
	  &src, 
	  &dst,
	  &value);
  PRINTF("set-loss %d->%d %d\n", 
	 src, 
	 dst, 
	 value);
  /* sendMessage(TESTBED_SET_LOSS_RATE,
     src,
     dst,
     value);*/
  return 0;
}

/* */
static int getStats(int argc,
		     char * argv[]) {
  /*  int   src, dst;
      char *s = argv[0];*/
  
  /* sscanf (s, "%d %d", &src, &dst); */
  /*
  PRINTF("get-trust %d->%d\n", src, dst);  
  if (OK == sendMessage(TESTBED_GET_TVALUE, src, dst, 0)) {
    CS_HEADER * hdr = NULL;

    readFromSocket(nodes[src].sock, (CS_HEADER **)&hdr);
    intptr = (int *)hdr->data;
    value = htonl(*intptr);
    PRINTF("Trust value between %d and %d: %d\n", src, dst, value);
    FREE(hdr);
  }
*/
  return 42;
}


static int printOnlineHelp(int argc,
			   char * argv[]) {
  PRINTF("nodes filename          ; load nodes in the network from file\n"\
	 "connections filename    ; load the connections among the nodes from file\n"\
	 "delete filename         ; load delete information from file\n"\
	 "showups                 ; show the nodes that are up\n"\
	 "add src dst             ; add a connection from node src to dst\n"\
	 "del src dst             ; delete a connection from node src to dst\n"\
	 "set-trust src dst value ; set value as the trust value in src for dst\n"\
	 "get-trust src dst       ; get trust value in src for dst\n"\
	 "set-bw src dst in out   ; set in and out bandwidth between src and dst\n"
	 "set-loss src dst loss   ; set the loss probability between src and dst\n"
	 "get-stats peer          ; get peer statistics\n"
	 );
  return 0;
}

static int showUps(int argc,
		   char * argv[]) {
  int i;
  for (i = 0; i < nnodes; i++)
    PRINTF("%d - %s:%d\n", 
	   i+1, 
	   nodes[i].ips, 
	   nodes[i].port);
  return 0;
}

static int doExit(int argc,
		  char * argv[]) {
  do_quit = YES;
  return 0;
}

/* exported symbols */

/* ****************** command set ****************** */

CMD_ENTRY commands[] = {
  { "help", &printOnlineHelp },
  { "get-trust", &getTrust },
  { "set-bw", &setBW },
  { "set-trust", &setTrust },
  { "add-node", &addNode },  
  { "add", &addConnection },
  { "del", &delConnection },
  { "exit", &doExit },
  { "showups", &showUps}, 
  { "set-loss", &setLoss} ,
  { "get-stats", &getStats}, 
  { "load-module", &loadModule },
  { "unload-module", &unloadModule },
  { NULL, NULL }, /* termination */
};

/**
 * Should the driver exit?
 */
int do_quit = NO;



/* end of commands.c */
