/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/testbed/gnunet-testbed.c 
 * @brief A Testbed tool for performing distributed experiments
 * @author Ronaldo Alves Ferreira
 * @author Christian Grothoff
 * @author Murali Krishan Ramanathan
 *
 * Todo:
 * - extend "alias"es in bash to contain an argument that specifies
 *   which GNUnet configuration file should be used.
 **/

#include "testbed.h"
#include "socket.h"
#include "commands.h"
#include <sys/wait.h>

#define TESTBED_VERSION		"0.0.2"
#define HELPER                  "==HELPER=="

/* we may want to change SHELL and PORT into values
   obtained from the configuration... */

#define SHELL                   (NULL == getenv("BASH") ? "/bin/bash" : getenv("BASH"))

#define PORT 2088

/* TB_ALIASES should probably be forced to live somewhere
   under ~/.gnunet */
#define TB_ALIASES "/tmp/gnunet-testbed-aliasdefinitions"


/**
 * Name of gnunet-testbed binary.
 */
static char * testbedArg0;


/* ****************** scriptability *************** */

/**
 * This is the main method of the helper-process that
 * is invoked from the bash-process.  Helper encapsulates
 * the command from the shell in a stream, sends it
 * over the socket to the main gnunet-testbed process,
 * retrieves there result and outputs he result back
 * to bash.
 */
static int helper_main(int argc,
		       char * argv[]) {
  int retVal;
  int i;
  unsigned int len;
  char * buf;
  int res;
  struct sockaddr_in soaddr;

  if (argc == 1) {
    fprintf(stderr,
	    "ERROR: must have at least one argument!\n");
    return -1;
  }

  sock = socket(PF_INET, SOCK_STREAM, 6); /* 6: TCP */
  if (sock == -1) {
    LOG(LOG_FAILURE,
	"FAILURE: Can not create socket (%s).\n",
	strerror(errno));
    return SYSERR;
  }
  soaddr.sin_family = AF_INET;
  soaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  soaddr.sin_port = htons(PORT);
  res = connect(sock, 
		(struct sockaddr*)&soaddr,
		sizeof(soaddr));
  if ( (res < 0) &&
       (errno != EINPROGRESS) ) {
    LOG(LOG_INFO,
	"INFO: tcpio: Can not connect to LOOPBACK:%d (%s)\n",
	PORT,
	strerror(errno));
    CLOSE(sock);
    sock = -1;
    return SYSERR;
  }

  /* write command to socket */
  socketSend(strlen(argv[1]),
	   SOCKET_BEGIN_COMMAND,
	   argv[1]);
  /* write args to socket */
  for (i=2;i<argc;i++)
    socketSend(strlen(argv[i]),
	     SOCKET_ADD_ARGUMENT,
	     argv[i]);
  socketSend(0, 
	   SOCKET_END_COMMAND,
	   NULL);

  /* read result from socket, print to stderr,
     obtain retVal */
  i = SOCKET_PRINTF;
  buf = NULL;
  while (i == SOCKET_PRINTF) {
    FREENONNULL(buf);
    buf = NULL;
    i = readSocket(&buf, &len);
    if (i == SOCKET_PRINTF)
      fprintf(stdout,
	      "%.*s", len, buf);
  }  
  retVal = *(int*)buf;
  FREE(buf);  
  close(sock);
  return retVal;
}

/**
 * This is the "bash" process.  Execs bash.
 * @returns never
 */
static void bash_main() {
  FILE * aliases;
  char * argv[] = {
    NULL, /* replaced by SHELL */
    "--init-file",
    TB_ALIASES,
    "-i",
    NULL,
  };
  int i;
  
  argv[0] = SHELL;
  aliases = fopen(TB_ALIASES,
		  "w+");
  fprintf(aliases,
	  "export PS1=\"$PS1 @GTB@ \"\n");
  i=0;
  while (commands[i].command != NULL) {
    if (0 == strcmp("exit", commands[i].command)) {
      fprintf(aliases,
	      "alias exit=\"%s ==HELPER== exit ; exit\"\n",
	      testbedArg0);
    } else {
      fprintf(aliases,
	      "alias %s=\"%s ==HELPER== %s\"\n",
	      commands[i].command,
	      testbedArg0,
	      commands[i].command);
    }
    i++;
  }
  fclose(aliases);
  execvp(SHELL,
	 argv);       
  fprintf(stderr,
	  "FATAL: could not execute %s: %s\n",
	  SHELL,
	  strerror(errno));
}

/**
 * This is the main method of the server.  It reads
 * commands from the socket that are created by helper process.
 * It is the main process that also keeps the state
 * throughout the session.
 *
 * @param bash_pid the process ID of the child that is bash
 */
static int server_main(pid_t bash_pid) {
  int status;  
  int i;
  int ssock;
  int lenOfIncomingAddr;
  struct sockaddr_in serverAddr, clientAddr;
  int secs = 5;
  const int on = 1;

  /* create the socket */
 CREATE_SOCKET:
  while ( (ssock = socket(PF_INET, 
			  SOCK_STREAM, 
			  0)) < 0) {
    LOG(LOG_ERROR, 
	"ERROR opening socket (%s). No client service started. Trying again in 30 seconds.\n",
	strerror(errno));
    sleep(30);
  }

  serverAddr.sin_family = AF_INET;
  serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  serverAddr.sin_port = htons(PORT);
 
  if ( setsockopt(ssock, 
		  SOL_SOCKET, 
		  SO_REUSEADDR, 
		  &on, sizeof(on)) < 0 )
    perror("setsockopt");

  /* bind the socket */
  if (bind(ssock, 
	   (struct sockaddr *) &serverAddr,
	   sizeof(serverAddr)) < 0) {
    LOG(LOG_ERROR, 
	"ERROR (%s) binding the TCP listener to port %d. No proxy service started.\nTrying again in %d seconds...\n",
	strerror(errno),
	PORT, 
	secs);
    sleep(secs);
    secs += 5; /* slow progression... */
    CLOSE(ssock);
    goto CREATE_SOCKET;
  }

  do_quit = NO;
  /* FIXME:
     add handler for SIGCHLD and break the loop
     if the child did exit (e.g. with CTRL-D)! */
  listen(ssock, 5);
  while ( (do_quit == NO) &&
	  (0 == waitpid(bash_pid, 
			&status, 
			WNOHANG)) ) {
    char * command;
    char ** args;
    int argc;
    char * buf;
    unsigned int len;

    lenOfIncomingAddr = sizeof(clientAddr); 
    sock = accept(ssock,
		  (struct sockaddr *)&clientAddr, 
		  &lenOfIncomingAddr);
    if (sock < 0)
      continue;
    /* FIXME: add access control here! */
    /* read from socket, run reaction,
       return result value and stdout value;
       possibly set doExit to 1 to exit */    
    buf = NULL;
    if (SOCKET_BEGIN_COMMAND != readSocket(&buf,
				       &len)) {
      fprintf(stderr,
	      "FATAL: protocol violation on socket.  Expected command.\n");
      return -1;
    }
    command = MALLOC(len+1);
    memcpy(command, buf, len);
    command[len] = '\0';    
    argc = 0;
    args = NULL;
    FREE(buf);
    buf = NULL;
    while (SOCKET_ADD_ARGUMENT == readSocket(&buf, 
					 &len)) {
      GROW(args,
	   argc,
	   argc+1);
      args[argc-1] = MALLOC(len+1);
      memcpy(args[argc-1],
	     buf,
	     len);	     
      args[argc-1][len] = '\0';
      FREE(buf);
      buf = NULL;
    }
    i = 0;
    while (commands[i].command != NULL) {
      if (0 == strcmp(commands[i].command,
		      command)) {
	int ret;
	ret = commands[i].handler(argc, args);
	socketSend(sizeof(unsigned int),
		 SOCKET_RETVAL,
		 &ret);
	break;
      }
      i++;
    }
    if (commands[i].command == NULL) {
      /* should never happen unless the user
	 plays by hand with the aliases... */
      i = -1;
      PRINTF("command %s not found!\n",
	     command);
      socketSend(sizeof(unsigned int),
	       SOCKET_RETVAL,
	       &i);
    }    
    close(sock);
    sock = -1;
  }
  /* just to be certain, we could have
     left the main loop due to doExit... */
  waitpid(bash_pid, 
	  &status, 
	   WNOHANG);
  return status;
}


/* *************** command line options *********** */

/* */
static void testbedUsage() {
  /* no non-standard arguments at this point */
}

/**
 * Parse the options, set the timeout.
 *
 * @param argc the number of options
 * @param argv the option list (including keywords)
 * @return OK on error, SYSERR if we should exit 
 **/
static int parseOptions(int argc,
			char ** argv) {
  int option_index;
  int c;  
  
  FREENONNULL(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL));
  while (1) {
    static struct GNoption long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { 0,0,0,0 }
    };    
   
    option_index = 0;
    c = GNgetopt_long(argc,
		      argv, 
		      "vhdc:L:", 
		      long_options, 
		      &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, GNoptarg))
      continue;
    switch(c) {
    case 'v': 
      printf("GNUnet v%s, gnunet-testbed v%s\n",
	     VERSION,
	     TESTBED_VERSION);
      return SYSERR;

    case 'h': 
      printf("USAGE: gnunet-testbed\n");
      printHelpDefault();
      testbedUsage();
      return SYSERR;
    default: 
      LOG(LOG_FAILURE,
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      return -1;
    } /* end of parsing commandline */
  } /* while (1) */
  return OK;
}

/* **************** main **************** */

/**
 * Create a testbed environment with multiple peers to collect 
 * GNUnet statistics.  Note that the same binary is used for
 * two distinct purposes.  One is the stateful master process
 * that maintains the testbed.  A second type of process is
 * created if HELPER is the first argument.  These processes
 * are created by the user from the SHELL and they merely serve
 * to communicate back the user-commands to the master-process.
 *
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return 0 on success, -1 on error
 **/   
int main(int argc, 
	 char ** argv) {
  pid_t pid;

  testbedArg0 = argv[0];
  /* first, evaluate if we are the special
     helper process.  If so, eat HELPER
     argument and run helper_main */
  if (argc > 1) {
    if (strcmp(argv[1], HELPER) == 0) {
      argv[1] = argv[0];
      return helper_main(argc -1,
			 &argv[1]);
    }
  }

  if (SYSERR == initUtil(argc, argv, &parseOptions))
    return -1; 

  /* we are the main testbed process.  Fork of
     a shell and start processing from the socket */

  pid = fork();
  if (pid < 0)
    errexit("FATAL: fork failed: %s\n",
	    strerror(errno));
  if (pid == 0) {
    doneUtil(); 
    bash_main();
    return 0; /* unreached */
  } else {
    int ret;
    /*fclose(stdin);
      fclose(stdout);*/
    /* run actual main loop */
    ret = server_main(pid);
    /* just to be certain */
    kill(pid, SIGHUP);
    /* proper shutdown... */
    doneUtil();
    unlink(TB_ALIASES);
    return ret;
  } 
}

/* end of gnunet-testbed.c */ 
