/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file test/hk_gcrypt.c
 * @author Christian Grothoff
 **/

#include "util/xmalloc.h"
#include "util/returnvalues.h"
#include "util/logging.h"
#include "util/cron.h"
#include "util/hashing.h"
#include <netinet/in.h>
#include "hk_gcrypt.h"

#define HOSTKEY_LEN 2048
#define EXTRA_CHECKS YES

/**
 * If target != size, move target bytes to the
 * end of the size-sized buffer and zero out the
 * first target-size bytes.
 **/
static void adjust(char * buf,
		   int size,
		   int target) {
  if (size < target) {
    memmove(&buf[target-size],
	    buf,
	    size);    
    memset(buf, 
	   0, 
	   target-size);
  }
}

/**
 * Initialize Random number generator.
 **/
void ginitRAND() {
  gcry_control(GCRYCTL_DISABLE_SECMEM, 0);
  if (! gcry_check_version(GCRYPT_VERSION))
    errexit("FATAL: libgcrypt has not the exptected version %s\n",
	    GCRYPT_VERSION);
  srand((unsigned int)time(NULL));
}


/**
 * This HostKey implementation uses RSA.
 **/
GCRY_SEXP gmakeHostkey() {
  GCRY_SEXP s_key;
  GCRY_SEXP s_keyparam;
  int rc;

  rc = gcry_sexp_build(&s_keyparam, 
		       NULL, 
		       "(genkey(rsa(nbits %d)(rsa-use-e 3:257)))",
		       HOSTKEY_LEN);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: failed to convert keyparam: %s\n", 
	gcry_strerror (rc));
    return NULL;
  }
  rc = gcry_pk_genkey(&s_key,
		      s_keyparam);
  gcry_sexp_release(s_keyparam);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: key generation failed: %s\n", 
	gcry_strerror (rc));
    return NULL;
  }

#if EXTRA_CHECKS
  if (gcry_pk_testkey(s_key)) {
    LOG(LOG_ERROR,
	"ERROR: freshly created hostkey is not sane!\n");
    return NULL;
  }
#endif
  return s_key;
}

/**
 * Free memory occupied by hostkey
 **/
void gfreeHostkey(GCRY_SEXP hostkey) {
  gcry_sexp_release(hostkey);
}

static int key_from_sexp( GCRY_MPI *array,
			  GCRY_SEXP sexp, 
			  const char *topname, 
			  const char *elems ) {
  GCRY_SEXP list, l2;
  const char *s;
  int i, idx;
  
  list = gcry_sexp_find_token( sexp, topname, 0 );
  if( !list )
    return GCRYERR_INV_OBJ;
  l2 = gcry_sexp_cadr( list );
  gcry_sexp_release ( list );
  list = l2;
  if( !list )
    return GCRYERR_NO_OBJ;
  
  idx = 0;
  for(s=elems; *s; s++, idx++ ) {
    l2 = gcry_sexp_find_token( list, s, 1 );
    if( !l2 ) {
      for(i=0; i<idx; i++) {
	gcry_free( array[i] );
	array[i] = NULL;
      }
      gcry_sexp_release ( list );
      return GCRYERR_NO_OBJ; /* required parameter not found */
    }
    array[idx] = gcry_sexp_nth_mpi( l2, 1, GCRYMPI_FMT_USG );
    gcry_sexp_release ( l2 );
    if( !array[idx] ) {
      for(i=0; i<idx; i++) {
	gcry_free( array[i] );
	array[i] = NULL;
      }
      gcry_sexp_release ( list );
      return GCRYERR_INV_OBJ; /* required parameter is invalid */
    }
  }
  gcry_sexp_release ( list );  
  return 0;
}

/**
 * Extract the public key of the host.
 * @param hostkey the hostkey to extract into the result.
 * @param result where to write the result.
 **/
void ggetPublicKey(GCRY_SEXP hostkey,
		  PublicKey * result) {
  GcryMPI skey[2];
  int size;
  int rc;
  
  rc = key_from_sexp(skey, 
		     hostkey, 
		     "public-key", 
		     "ne");
  if (rc)
    rc = key_from_sexp(skey, 
		       hostkey, 
		       "private-key", 
		       "ne");    
  if (rc)
    rc = key_from_sexp(skey, 
		       hostkey, 
		       "rsa", 
		       "ne");    
  if (rc) 
    errexit("FATAL: getPublicKey: key_from_sexp failed: %s\n", 
	    gcry_strerror(rc));
  
  result->len = htons(sizeof(PublicKey) - sizeof(result->padding));
  result->sizen = htons(RSA_ENC_LEN);
  result->padding = 0;
  size = RSA_ENC_LEN;
  rc = gcry_mpi_print(GCRYMPI_FMT_USG, 
		      &result->key[0], 
		      &size, 
		      skey[0]);
  if (rc) 
    errexit("FATAL: gcry_mpi_print of n failed: %s\n",
	    gcry_strerror(rc));
  adjust(&result->key[0], size, RSA_ENC_LEN);

  size = RSA_KEY_LEN - RSA_ENC_LEN; 
  rc = gcry_mpi_print(GCRYMPI_FMT_USG, 
		      &result->key[RSA_ENC_LEN], 
		      &size, 
		      skey[1]);
  if (rc) 
    errexit("FATAL: gcry_mpi_print of e failed: %s\n",
	    gcry_strerror(rc));
   adjust(&result->key[RSA_ENC_LEN], 
	 size,
	 RSA_KEY_LEN - RSA_ENC_LEN);
}


/**
 * Internal: publicKey => RSA-Key
 **/
static GCRY_SEXP public2Hostkey(PublicKey * publicKey) {
  GCRY_SEXP result;
  GcryMPI n;
  GcryMPI e;
  size_t size;
  size_t erroff;
  int rc;
  // len = sizen + sizee + 4 == publickey - padding!
  if ( ( ntohs(publicKey->sizen) != RSA_ENC_LEN ) ||
       ( ntohs(publicKey->len) != sizeof(PublicKey) - sizeof(publicKey->padding)) ) {
    LOG(LOG_ERROR,
	"ERROR: public2Hostkey: received invalid publicKey (%d, %d)\n",
	ntohs(publicKey->len),
	ntohs(publicKey->sizen)); 
    return NULL;
  }
  size = RSA_ENC_LEN;
  rc = gcry_mpi_scan(&n,
		     GCRYMPI_FMT_USG,
		     &publicKey->key[0],
		     &size);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: gcry_mpi_scan of n failed (%s)\n",
	gcry_strerror(rc));
    return NULL;
  }
  size = RSA_KEY_LEN - RSA_ENC_LEN;
  rc = gcry_mpi_scan(&e,
		     GCRYMPI_FMT_USG,
		     &publicKey->key[RSA_ENC_LEN],
		     &size);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: gcry_mpi_scan of e failed (%s)\n",
	gcry_strerror(rc));
    gcry_mpi_release(n);
    return NULL;
  }
  rc = gcry_sexp_build(&result, 
		       &erroff,
 		       "(public-key(rsa(n %m)(e %m)))",
		       n,
		       e);
  gcry_mpi_release(n);
  gcry_mpi_release(e);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: gcry_sexp_build of public key failed (%s at position %d)\n",
	gcry_strerror(rc),
	erroff);
    return NULL;
  }  
  return result;
}

/**
 * Encode the private key in a format suitable for
 * storing it into a file.
 * @returns encoding of the private key.
 *    The first 4 bytes give the size of the array, as usual.
 **/
HostKeyEncoded * gencodeHostkey(GCRY_SEXP hostkey) {
  /* libgcrypt */

  HostKeyEncoded * retval;
  GcryMPI pkv[6];
  void * pbu[6];
  size_t sizes[6];
  int rc;
  int i;
  int size;

#if EXTRA_CHECKS
  if (gcry_pk_testkey(hostkey)) {
    LOG(LOG_ERROR,
	"ERROR: encodeHostkey: hostkey is not sane!\n");
    return NULL;
  }
#endif

  memset(pkv, 0, sizeof(GcryMPI) * 6);
  rc = key_from_sexp(pkv,
		     hostkey,
		     "private-key",
		     "nedpqu");
  if (rc)
    rc = key_from_sexp(pkv,
		       hostkey,
		       "rsa",
		       "nedpqu");
  if (rc) 
    rc = key_from_sexp(pkv,
		       hostkey,
		       "private-key",
		       "nedpq");
  if (rc)
    rc = key_from_sexp(pkv,
		       hostkey,
		       "rsa",
		       "nedpq");
  if (rc) 
    rc = key_from_sexp(pkv,
		       hostkey,
		       "private-key",
		       "ned");
  if (rc)
    rc = key_from_sexp(pkv,
		       hostkey,
		       "rsa",
		       "ned");
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encodeHostkey: key_from_sexp failed: %s\n", 
	gcry_strerror (rc));
    return NULL;
  }
  size = sizeof(HostKeyEncoded);
  for (i=0;i<6;i++) {
    if (pkv[i] != NULL) {
      rc = gcry_mpi_aprint(GCRYMPI_FMT_USG,
			   &pbu[i],
			   &sizes[i],
			   pkv[i]);
      size += sizes[i];
      if (rc) {
	LOG(LOG_ERROR,
	    "ERROR: gcry_mpi_aprint failed: %s\n", 
	    gcry_strerror (rc));
	while (i>0) {
	  FREENONNULL(pbu[--i]);
	}
	for (i=0;i<6;i++)
	  if (pkv[i] != NULL)
	    gcry_mpi_release(pkv[i]);
	return NULL;
      }
    } else {
      pbu[i] = NULL;
      sizes[i] = 0;
    }
  }
  if (size >= 65536) 
    errexit("FATAL: size of serialized private key >= 64k\n");  
  retval = MALLOC(size);
  retval->len = htons(size);
  i = 0;
  retval->sizen = htons(sizes[0]);
  memcpy(&retval->key[i], 
	 pbu[0],
	 sizes[0]);
  i += sizes[0];
  retval->sizee = htons(sizes[1]);
  memcpy(&retval->key[i], 
	 pbu[1],
	 sizes[1]);
  i += sizes[1];
  retval->sized = htons(sizes[2]);
  memcpy(&retval->key[i], 
	 pbu[2],
	 sizes[2]);
  i += sizes[2];
  retval->sizep = htons(sizes[4]);
  memcpy(&retval->key[i], 
	 pbu[4],
	 sizes[4]);
  i += sizes[4];
  retval->sizeq = htons(sizes[3]);
  memcpy(&retval->key[i], 
	 pbu[3],
	 sizes[3]);
  i += sizes[3];
  retval->sizedmp1 = htons(0);
  retval->sizedmq1 = htons(0);
  memcpy(&retval->key[i], 
	 pbu[5],
	 sizes[5]);
  for (i=0;i<6;i++) {
    if (pkv[i] != NULL)
      gcry_mpi_release(pkv[i]);
    FREENONNULL(pbu[i]);
  }
  return retval;
}

/**
 * Decode the private key from the file-format back
 * to the "normal", internal format.
 **/
GCRY_SEXP gdecodeHostkey(HostKeyEncoded * encoding) {
  GCRY_SEXP res;
  GcryMPI n,e,d,p,q,u;
  int rc;
  size_t size;
  int pos;

  pos = 0;
  size = ntohs(encoding->sizen);
  rc = gcry_mpi_scan(&n,
		     GCRYMPI_FMT_USG,
		     &encoding->key[pos],
		     &size);
  pos += ntohs(encoding->sizen);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: could not decode hostkey (%s)\n",
	gcry_strerror(rc));
    return NULL;
  }
  size = ntohs(encoding->sizee);
  rc = gcry_mpi_scan(&e,
		     GCRYMPI_FMT_USG,
		     &encoding->key[pos],
		     &size);
  pos += ntohs(encoding->sizee);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: could not decode hostkey (%s)\n",
	gcry_strerror(rc));
    gcry_mpi_release(n);
    return NULL;
  }
  size = ntohs(encoding->sized);
  rc = gcry_mpi_scan(&d,
		     GCRYMPI_FMT_USG,
		     &encoding->key[pos],
		     &size);
  pos += ntohs(encoding->sized);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: could not decode hostkey (%s)\n",
	gcry_strerror(rc));
    gcry_mpi_release(n);
    gcry_mpi_release(e);
    return NULL;
  }
  size = ntohs(encoding->sizep);
  if (size > 0) {
    rc = gcry_mpi_scan(&q,
		       GCRYMPI_FMT_USG,
		       &encoding->key[pos],
		       &size);
    pos += ntohs(encoding->sizep);
    if (rc) {
      LOG(LOG_ERROR,
	  "ERROR: could not decode hostkey (%s)\n",
	  gcry_strerror(rc));
      gcry_mpi_release(n);
      gcry_mpi_release(e);
      gcry_mpi_release(d);
      return NULL;
    }
  } else
    q = NULL;
  size = ntohs(encoding->sizeq);
  if (size > 0) {
    rc = gcry_mpi_scan(&p,
		       GCRYMPI_FMT_USG,
		       &encoding->key[pos],
		       &size);
    pos += ntohs(encoding->sizeq);
    if (rc) {
      LOG(LOG_ERROR,
	  "ERROR: could not decode hostkey (%s)\n",
	  gcry_strerror(rc));
      gcry_mpi_release(n);
      gcry_mpi_release(e);
      gcry_mpi_release(d);
      if (q != NULL)
	gcry_mpi_release(q);
      return NULL;
    }
  } else
    p = NULL;
  pos += ntohs(encoding->sizedmp1);
  pos += ntohs(encoding->sizedmq1);
  
  size = ntohs(encoding->len) - sizeof(HostKeyEncoded) - pos;
  if (size > 0) {
    rc = gcry_mpi_scan(&u,
		       GCRYMPI_FMT_USG,
		       &encoding->key[pos],
		       &size);
    if (rc) {
      LOG(LOG_ERROR,
	  "ERROR: could not decode hostkey (%s)\n",
	  gcry_strerror(rc));
      gcry_mpi_release(n);
      gcry_mpi_release(e);
      gcry_mpi_release(d);
      if (p != NULL)
	gcry_mpi_release(p);
      if (q != NULL)
	gcry_mpi_release(q);
      return NULL;
    }
  } else
    u = NULL;
  if ( (p != NULL) &&
       (q != NULL) &&
       (u != NULL) ) {
    rc = gcry_sexp_build(&res,
			 &size, /* erroff */
			 "(private-key(rsa(n %m)(e %m)(d %m)(p %m)(q %m)(u %m)))",
			 n, e, d, p, q, u);
  } else {
    if ( (p != NULL) &&
	 (q != NULL) ) {
      rc = gcry_sexp_build(&res,
			   &size, /* erroff */
			   "(private-key(rsa(n %m)(e %m)(d %m)(p %m)(q %m)))",
			   n, e, d, p, q);
    } else {
      rc = gcry_sexp_build(&res,
			   &size, /* erroff */
			   "(private-key(rsa(n %m)(e %m)(d %m)))",
			   n, e, d);
    }
  }
  gcry_mpi_release(n);
  gcry_mpi_release(e);
  gcry_mpi_release(d);
  if (p != NULL)
    gcry_mpi_release(p);
  if (q != NULL)
    gcry_mpi_release(q);
  if (u != NULL)
    gcry_mpi_release(u);

  if (rc) {
     LOG(LOG_ERROR,
	  "ERROR: could not decode hostkey (%s at %d)\n",
	  gcry_strerror(rc),
	 size);
  }
#if EXTRA_CHECKS
  if (gcry_pk_testkey(res)) {
    LOG(LOG_ERROR,
	"ERROR: decodeHostkey: hostkey is not sane!\n");
    return NULL;
  }
#endif
 return res;
}

/**
 * Encrypt a block with the public key of another host that uses the
 * same cyper.
 *
 * @param block the block to encrypt
 * @param size the size of block
 * @param publicKey the encoded public key used to encrypt
 * @param target where to store the encrypted block
 * @returns SYSERR on error, OK if ok
 **/
int gencryptHostkey(void * block, 
		   unsigned short size,
		   PublicKey * publicKey,
		   RSAEncryptedData * target) {
  /* libgcrypt */
  GcrySexp result;
  GcrySexp data;
  GcrySexp pubkey;
  GcryMPI val;
  GcryMPI rval;
  size_t isize;
  size_t erroff;
  int rc;
  
  pubkey = public2Hostkey(publicKey);
  isize = size;
  rc = gcry_mpi_scan(&val,
		     GCRYMPI_FMT_USG,
		     block,
		     &isize);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encryptHostkey - gcry_mpi_scan failed (%s)\n", 
	gcry_strerror (rc));
    gfreeHostkey(pubkey);
    return SYSERR;
  }
  rc = gcry_sexp_build(&data,
		       &erroff,
		       "(data (flags pkcs1)(value %m))",
		       val);
  gcry_mpi_release(val);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encryptHostkey - gcry_sexp_build failed (%s at offset %d)\n", 
	gcry_strerror (rc),
	erroff);
    gfreeHostkey(pubkey);
    return SYSERR;
  }
  
  rc = gcry_pk_encrypt(&result, data, pubkey);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encryptHostkey - gcry_pk_encrypt failed (%s)\n", 
	gcry_strerror (rc),
	erroff);
    gcry_sexp_release(data);
    gfreeHostkey(pubkey);
    return SYSERR;
  }
  gcry_sexp_release(data);
  gfreeHostkey(pubkey);

  rc = key_from_sexp(&rval,
		     result,
		     "rsa",
		     "a");
  gcry_sexp_release(result);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: hostkeyEncrypt: key_from_sexp failed (%d, %s)\n",
	rc, 
	gcry_strerror(rc));
    return SYSERR;
  }
  isize = sizeof(RSAEncryptedData);
  rc = gcry_mpi_print(GCRYMPI_FMT_USG,
		      (char*)target,
		      &isize, 
		      rval);
  gcry_mpi_release(rval);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encryptHostkey - gcry_mpi_print failed (%s)\n", 
	gcry_strerror (rc));
    return SYSERR;
  }
  adjust(&target->encoding[0],
	 isize,
	 sizeof(RSAEncryptedData));
  return OK;
}

/**
 * Decrypt a given block with the hostkey. 
 *
 * @param hostkey the hostkey with which to decrypt this block
 * @param block the data to decrypt, encoded as returned by encrypt
 * @param result pointer to a location where the result can be stored
 * @param max the maximum number of bits to store for the result, if
 *        the decrypted block is bigger, an error is returned
 * @returns the size of the decrypted block, -1 on error
 **/
int gdecryptHostkey(GCRY_SEXP hostkey, 
		   RSAEncryptedData * block,
		   void * result,
		   unsigned int max) {
  GcrySexp resultsexp;
  GcrySexp data;
  size_t erroff;
  size_t size;
  GcryMPI val;
  int rc;
  char * endp;
  char * tmp;

#if EXTRA_CHECKS
  if (gcry_pk_testkey(hostkey)) {
    LOG(LOG_ERROR,
	"ERROR: decrypt: hostkey is not sane!\n");
    return -1;
  }
#endif
  size = sizeof(RSAEncryptedData);
  rc = gcry_mpi_scan(&val,
		     GCRYMPI_FMT_USG,
		     &block->encoding[0],
		     &size);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: hostkeyDecrypt: gcry_mpi_scan failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  rc = gcry_sexp_build(&data,
		       &erroff,
		       "(enc-val(flags)(rsa(a %m)))",
		       val);
  gcry_mpi_release(val);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: hostkeyDecrypt: gcry_sexp_build failed (%s at %d)\n",
	gcry_strerror(rc),
	erroff);
    return SYSERR;
  }
  rc = gcry_pk_decrypt(&resultsexp,
		       data,
		       hostkey);
  gcry_sexp_release(data);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: hostkeyDecrypt: gcry_pk_decrypt failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }

  /* resultsexp has format "(value %m)" */
  val = gcry_sexp_nth_mpi(resultsexp, 1, GCRYMPI_FMT_USG);
  gcry_sexp_release(resultsexp);
  if (val == NULL) {
    LOG(LOG_ERROR,
	"ERROR: gcry_sexp_nth_mpi failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  tmp = MALLOC(max + HOSTKEY_LEN/8);
  size = max+HOSTKEY_LEN/8;
  rc = gcry_mpi_print(GCRYMPI_FMT_USG,
		      tmp,
		      &size,
		      val);
  gcry_mpi_release(val);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: gcry_mpi_print failed (%s)\n",
	gcry_strerror(rc));
    FREE(tmp);
    return SYSERR;
  }

  endp = tmp;
  if (*endp == 0) {
    endp++;
    size--;
  }
  if ( (!size) || 
       (*endp != 0x02 )) {
    LOG(LOG_ERROR,
	"ERROR: not a pkcs-1 block type 2 (size=%d, *tmp=%d)!\n",
	size, (int)*tmp);
    FREE(tmp);
    return SYSERR;
  }
  /* serach the end of the padding */
  while ( (size > 0) && 
	  ((*endp) != 0) ) {
    size--;
    endp++;
  }
  if ( (size == 0) || 
       (*endp != 0x0) ) {
    LOG(LOG_ERROR,
	"ERROR: not a pkcs-1 block type 2 (size=%d, *endp=%d)\n",
	size, (int) *endp);
    FREE(tmp);
    return SYSERR;
  }
  size--;
  endp++;

  if (size > max)
    size = max;
  memcpy(result,
	 endp,
	 size);
  FREE(tmp);
  return size;
}

/**
 * Sign a given block.
 *
 * @param hostkey the hostkey with which to sign this block
 * @param size how many bytes to sign
 * @param block the data to sign
 * @param sig where to write the signature
 * @return SYSERR on error, OK on success
 **/
int gsign(GCRY_SEXP hostkey, 
	 unsigned short size,
	 void * block,
	 Signature * sig) {
  GcrySexp result;
  GcrySexp data;
  size_t ssize;
  GcryMPI rval;
  HashCode160 hc;
  char * buff;
  int bufSize;
  int rc;

  hash(block, size, &hc);
#define FORMATSTRING "(4:data(5:flags5:pkcs1)(4:hash6:rmd16020:01234567890123456789))"
  bufSize = strlen(FORMATSTRING) + 1;
  buff = MALLOC(bufSize);
  memcpy(buff,
	 FORMATSTRING,
	 bufSize);
  memcpy(&buff[bufSize - strlen("012345678901234567890))")],
	 &hc,
	 sizeof(HashCode160));
  rc = gcry_sexp_new(&data,
		     buff,
		     bufSize, 
		     0);
  FREE(buff);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: gcry_sexp_new failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  rc = gcry_pk_sign(&result, data, hostkey);
  gcry_sexp_release(data);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: gcry_pk_sign failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  rc = key_from_sexp(&rval,
		     result,
		     "rsa", 
		     "s");
  gcry_sexp_release(result);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: key_from_sexp failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  ssize = sizeof(Signature);
  rc = gcry_mpi_print(GCRYMPI_FMT_USG,
		      (char*)sig,
		      &ssize,
		      rval);
  gcry_mpi_release(rval);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: gcry_mpi_print failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  adjust(&sig->sig[0], 
	 ssize,
	 sizeof(Signature));
  return OK;
}

/**
 * Verify signature.
 *
 * @param block the signed data
 * @param len the length of the block 
 * @param sig signature
 * @param publicKey public key of the signer
 * @returns OK if ok, SYSERR if invalid
 **/
int gverifySig(void * block,
	      unsigned short len,
	      Signature * sig,	      
	      PublicKey * publicKey) {
  /* libgcrypt */
  GcrySexp data;
  GcrySexp sigdata;
  size_t size;
  GcryMPI val;
  GCRY_SEXP hostkey;
  HashCode160 hc;
  char * buff;
  int bufSize;
  size_t erroff;
  int rc;
 
  size = sizeof(Signature);
  rc = gcry_mpi_scan(&val,
		     GCRYMPI_FMT_USG,
		     (char*)sig,
		     &size);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: key_from_sexp failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  rc = gcry_sexp_build(&sigdata,
		       &erroff,
		       "(sig-val(rsa(s %m)))",
		       val);
  gcry_mpi_release(val);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: gcry_sexp_build failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }  
  hash(block, len, &hc);
  bufSize = strlen(FORMATSTRING) + 1;
  buff = MALLOC(bufSize);
  memcpy(buff,
	 FORMATSTRING,
	 bufSize);
  memcpy(&buff[strlen(FORMATSTRING) - strlen("01234567890123456789))")],
	 &hc,
	 sizeof(HashCode160));
  rc = gcry_sexp_new(&data,
		     buff,
		     bufSize, 
		     0);
  FREE(buff);
  hostkey = public2Hostkey(publicKey);
  rc = gcry_pk_verify(sigdata,
		      data,
		      hostkey);  
  gfreeHostkey(hostkey);
  gcry_sexp_release(data);
  gcry_sexp_release(sigdata);
  if (rc) {
    LOG(LOG_WARNING,
	"WARNING: signature verification failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  } else
    return OK;
}


/* end of hostkey.c */
