/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * @file util/configuration.c
 * @brief high-level configuration managment (with command-line overrides)
 * @author Christian Grothoff
 *
 * This file provides code to parse a configuration file and access
 * the data stored in it. It also provides methods to override options
 * which can be used to add command-line options to the configuration
 * API.
 **/

#include "gnunet_util.h"
#include "parseconfig.h"

/**
 * Type for user-defined configuration entries.
 **/
typedef struct UserConfStruct {
  char * section;
  char * option;
  char * stringValue;
  unsigned int intValue;
  struct UserConfStruct * next;
} UserConf;

/**
 * GNUnet configuration (OpenSSL datastructure)
 **/
static int parseConfigInit = NO;

/**
 * The filename of the config (for re-reading!)
 **/
static char * configuration_filename = NULL;

/**
 * Lock to access configuration (we may receive an
 * update signal at any time!)
 **/
static Mutex configLock;

/**
 * List of run-time options (not in the configuration
 * file, but for example from the command line).
 **/
static UserConf * userconfig = NULL;

/**
 * The command line strings (rest)
 **/ 
static char ** values;
static int valuesCount;

/**
 * Expand an expression of the form
 * "$FOO/BAR" to "DIRECTORY/BAR" where
 * either in the current section or
 * globally FOO is set to DIRECTORY.
 **/
static char * expandDollar(char * section,
			   char * orig) {
  int i;
  char * prefix;
  char * result;
  
  i=0;
  while ( (orig[i] != '/') &&
	  (orig[i] != '\0') )
    i++;
  if (orig[i] == '\0')
    return orig;
  orig[i] = '\0';
  prefix = getConfigurationString(section,
				  &orig[1]);
  if (prefix == NULL)
    prefix = getConfigurationString("", &orig[1]);
  if (prefix == NULL) {
    orig[i] = '/';
    return orig;
  }
  result = MALLOC(strlen(prefix) + 
		  strlen(&orig[i+1]) + 2);
  strcpy(result, prefix);
  strcat(result, "/");
  strcat(result, &orig[i+1]);
  FREE(prefix);
  FREE(orig);
  return result;
}

/**
 * Obtain a filename from the given section and option.  If the
 * filename is not specified, die with the given error message (do not
 * die if errMsg == NULL). 
 *
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 *
 * @param errMsg the errormessage, should contain two %s tokens for
 * the section and the option.
 *
 * @return the specified filename (caller must free), or NULL if no
 * filename was specified and errMsg == NULL
 **/
char * getFileName(char * section,
		   char * option,
		   const char * errMsg) {
  char * fn;  
  char * fnExpand;

  fn = getConfigurationString(section, 
			      option);
  if (fn == NULL) {
    if (errMsg == NULL)
      return NULL;
    else
      errexit(errMsg,
	      section,
	      option);
  }
  fnExpand = expandFileName(fn);   
  FREE(fn);
  return fnExpand;
}

/**
 * Read the configuration file.  The previous configuration will be
 * discarded if this method is invoked twice.
 **/
void readConfiguration() {
  /* getFileName aquires the mutex, so we better do this first */
  char * cfgName;
  char * expCfgName;

  cfgName = getConfigurationString("FILES",
				   "gnunet.conf");
  if (cfgName == NULL)
    expCfgName = expandFileName("~/.gnunet/gnunet.conf");
  else 
    expCfgName = expandFileName(cfgName);
  if (0 == assertIsFile(expCfgName)) {
    if (cfgName != NULL)
      errexit("Can not open configuration file %s\n",
	      cfgName);
    expCfgName = STRDUP("/etc/gnunet.conf");
    if (0 == assertIsFile(expCfgName)) 
      errexit("Can not find configuration at default locations %s or %s\n",
	      "~/.gnunet/gnunet.conf",
	      "/etc/gnunet.conf");    
  }
  FREENONNULL(cfgName);

  FREENONNULL(setConfigurationString("FILES",
				     "gnunet.conf",
				     expCfgName));
  MUTEX_LOCK(&configLock);
  FREENONNULL(configuration_filename);
  configuration_filename = expCfgName;
  
  if (parseConfigInit == YES) {
    doneParseConfig();
    parseConfigInit = NO;
  }
  if (0 != cfg_parse_file(configuration_filename))
    errexit("Bad configuration file %s.\n",
	    configuration_filename);
  parseConfigInit = YES;
  MUTEX_UNLOCK(&configLock);
}

static NotifyConfigurationUpdateCallback * cbl = NULL;
static int cbCnt = 0;

void registerConfigurationUpdateCallback
(NotifyConfigurationUpdateCallback cb) {
  MUTEX_LOCK(&configLock);
  GROW(cbl,
       cbCnt,
       cbCnt+1);
  cbl[cbCnt-1] = cb;
  MUTEX_UNLOCK(&configLock);
}

void unregisterConfigurationUpdateCallback
(NotifyConfigurationUpdateCallback cb) {
  int i;
  
  MUTEX_LOCK(&configLock);
  for (i=0;i<cbCnt;i++)
    if (cbl[i] == cb)
      break;
  if (i < cbCnt) {
    cbl[i] = cbl[cbCnt-1];
    GROW(cbl,
	 cbCnt,
	 cbCnt-1);
  } else 
    LOG(LOG_WARNING,
	"WARNING: unregisterConfigurationUpdateCallback called with handler %x which is not registered.\n",
	(int) cb);
  MUTEX_UNLOCK(&configLock);
}

void triggerGlobalConfigurationRefresh() {
  int i;
  MUTEX_LOCK(&configLock);
  for (i=0;i<cbCnt;i++)
    cbl[i]();
  MUTEX_UNLOCK(&configLock);
}

/**
 * This method must be called first! It reads
 * the config file and makes everything else
 * possible.
 **/
void initConfiguration() {
  MUTEX_CREATE_RECURSIVE(&configLock);
}

/**
 * This method may be called at last to clean up.
 * Afterwards everything but initConfiguration will result
 * in errors...
 **/
void doneConfiguration() {
  parseConfigInit = NO;
  doneParseConfig();
  FREENONNULL(configuration_filename);
  configuration_filename = NULL;
  MUTEX_DESTROY(&configLock);
  while (userconfig != NULL) {
    UserConf * tmp = userconfig;
    userconfig = userconfig->next;
    FREENONNULL(tmp->section);
    FREENONNULL(tmp->option);
    FREENONNULL(tmp->stringValue);
    FREE(tmp);
  }
}


/**
 * Obtain a string from the configuration.
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @return a freshly allocated string, caller must free!
 *   Note that the result can be NULL if the option is not set.
 **/
char * getConfigurationString(char * section,
			      char * option) {
  UserConf * pos;
  char * retval;

  if ( (section == NULL) || 
       (option == NULL) )
    errexit("getConfigurationString called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      if (pos->stringValue != NULL)
	retval = STRDUP(pos->stringValue);
      else
	retval = NULL; 
      MUTEX_UNLOCK(&configLock);
      if (retval != NULL)
	if (retval[0] == '$') {
	  retval = expandDollar(section,
				retval);
	}
      return retval;
    }
    pos = pos->next;
  }
  retval = NULL;
  if (parseConfigInit == YES)
    retval = cfg_get_str(section, option);
  if (retval != NULL)
    retval = STRDUP(retval);  
  MUTEX_UNLOCK(&configLock);
  if (retval != NULL)
    if (retval[0] == '$') 
      retval = expandDollar(section, retval);    
  return retval;
}

/**
 * Check if a string in the configuration matches a given value.  This
 * method should be preferred over getConfigurationString since this
 * method can avoid making a copy of the configuration string that
 * then must be freed by the caller.
 *
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @param value the value to compare against
 * @return YES or NO
 **/
int testConfigurationString(char * section,
			    char * option,
			    const char * value) {
  char * c;

  if ( (section == NULL) || 
       (option == NULL) )
    errexit("testConfigurationString called with section or option being NULL!\n");
  c = getConfigurationString(section, option);
  if (c == NULL) {
    if (value == NULL)
      return YES;
    else
      return NO;
  } else {
    int ret;
    if (value == NULL) {
      FREE(c);
      return NO;
    }
    if (0 == strcmp(c, value))
      ret = YES;
    else
      ret = NO;
    FREE(c);
    return ret;
  }
}

/**
 * Obtain an int from the configuration.
 *
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @return 0 if no option is specified
 **/
unsigned int getConfigurationInt(char * section,
				 char * option) {
  UserConf * pos;
  unsigned int retval;

  if ( (section == NULL) || (option == NULL) )
    errexit("getConfigurationInt called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      retval = pos->intValue;
      MUTEX_UNLOCK(&configLock);
      return retval;
    }
    pos = pos->next;
  }
  retval = 0;
  if (parseConfigInit == YES) 
    retval = cfg_get_signed_int(section, option);
  MUTEX_UNLOCK(&configLock);
  return retval;
}

/**
 * Set an option.
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @param value the value to use, may be NULL
 * @return the previous value (or NULL if none),
 *     caller must free!
 **/
char * setConfigurationString(char * section,
			      char * option,
			      const char * value) {
  UserConf * pos;
  UserConf * prev; 
  char * res;

  if ( (section == NULL) ||
       (option == NULL) )
    errexit("setConfigurationString called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  prev = NULL;
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      res = pos->stringValue;
      if (value != NULL)
	pos->stringValue = STRDUP(value);
      else
	pos->stringValue = NULL;
      MUTEX_UNLOCK(&configLock);
      return res;
    }
    prev = pos;
    pos = pos->next;
  }
  if (prev == NULL) {
    userconfig = MALLOC(sizeof(UserConf));
    pos = userconfig;
  } else {
    prev->next = MALLOC(sizeof(UserConf));
    pos = prev->next;
  }
  pos->section = STRDUP(section);
  pos->option = STRDUP(option);
  if (value != NULL)
    pos->stringValue = STRDUP(value);
  else 
    pos->stringValue = NULL;
  pos->intValue = 0;
  pos->next = NULL;
  res = NULL;
  if (parseConfigInit == YES) {
    res = cfg_get_str(section, option);
    if (res != NULL)
      res = STRDUP(res);
  }
  MUTEX_UNLOCK(&configLock);
  return res;    
}

/**
 * Set an option.
 * @param section from which section, may not be NULL
 * @param option which option, may not be NULL
 * @param value the value to use
 * @return the previous value (or 0 if none)
 **/
unsigned int setConfigurationInt(char * section,
				 char * option,
				 const unsigned int value) {
  UserConf * pos;
  UserConf * prev; 
  unsigned int res;

  if ( (section == NULL) || (option == NULL) )
    errexit("setConfigurationInt called with section or option being NULL!\n");
  MUTEX_LOCK(&configLock);
  prev = NULL;
  pos = userconfig;
  while (pos != NULL) {
    if ( (strcmp(section, pos->section) == 0) &&
	 (strcmp(option, pos->option) == 0) ) {
      res = pos->intValue;
      pos->intValue = value;
      MUTEX_UNLOCK(&configLock);
      return res;
    }
    prev = pos;
    pos = pos->next;
  }
  if (prev == NULL) {
    userconfig = MALLOC(sizeof(UserConf));
    pos = userconfig;
  } else {
    prev->next = MALLOC(sizeof(UserConf));
    pos = prev->next;
  }
  pos->section = STRDUP(section);
  pos->option = STRDUP(option);
  pos->stringValue = NULL;
  pos->intValue = value;
  pos->next = NULL;
  res = 0;
  if (parseConfigInit == YES)
    res = cfg_get_signed_int(section, option);
  MUTEX_UNLOCK(&configLock);
  return res; 
}

/**
 * Get the command line strings (the ones
 * remaining after getopt-style parsing).
 * @param value the values
 + @return the number of values
 **/
int getConfigurationStringList(char *** value) {
  char ** cpy;
  int i;

  cpy = MALLOC(sizeof(char*) * valuesCount);
  for (i=0;i<valuesCount;i++)
    cpy[i] = STRDUP(values[i]);
  *value = cpy;
  return valuesCount;
}

/**
 * Set the list of command line options (remainder
 * after getopt style parsing).
 * @param value the values 
 + @param count the number of values
 **/
void setConfigurationStringList(char ** value,
				int count) {
  values = value;
  valuesCount = count;
}



/* end of configuration.c */
